package com.jpattern.orm.generator.reflection;

import java.sql.ResultSet;
import java.util.Map;
import java.util.TreeMap;

import com.jpattern.orm.classmapper.IClassMapper;
import com.jpattern.orm.classmapper.IColumn;
import com.jpattern.orm.crud.AColumnValueGenerator;
import com.jpattern.orm.exception.OrmConfigurationException;
import com.jpattern.orm.exception.OrmReflectionException;
import com.jpattern.orm.generator.IOrmPersistor;

/**
 * A persistor implementation based on reflection
 * @author Francesco Cina'
 *
 * Mar 24, 2012
 */
public class ReflectionOrmPersistor<T> implements IOrmPersistor<T> {

	private final IClassMapper<T> classMapper;
	private final Map<String, FieldManipulator> manipulators = new TreeMap<String, FieldManipulator>();
	private VersionManipulator versionManipulator = new NullVersionManipulator();
	private GeneratorManipulator generatorManipulator = new NullGeneratorManipulator();

	public ReflectionOrmPersistor(IClassMapper<T> classMapper) throws OrmConfigurationException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		this.classMapper = classMapper;
		this.initManipulators();
		this.initVersionManipulator();
		this.initGeneratorManipulator();
	}

	private void initManipulators() throws OrmConfigurationException, SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
		for (final String columnJavaName : this.classMapper.getAllColumnJavaNames()) {
			final IColumn column = this.classMapper.getColumnWithJavaName(columnJavaName);
			final AGetFieldManipulator getFieldManipulator = AReflectionManipulatorFactory.getGetFieldManipulator(column);
			final ASetFieldManipulator setFieldManipulator = AReflectionManipulatorFactory.getSetFieldManipulator(column);
			this.manipulators.put(columnJavaName, new FieldManipulator(getFieldManipulator, setFieldManipulator, column.getName()));
		}
	}

	private void initVersionManipulator() throws OrmConfigurationException, SecurityException, NoSuchMethodException {
		if (this.classMapper.getTableMap().isVersionable()){
			final String javaColumnName = this.classMapper.getTableMap().getVersionJavaFieldName();
			final IColumn column = this.classMapper.getColumnWithJavaName(javaColumnName);
			final FieldManipulator manipulator = this.manipulators.get(javaColumnName);
			this.versionManipulator = AReflectionManipulatorFactory.getVersionManipulator(column, manipulator);
		}
	}

	private void initGeneratorManipulator() throws OrmConfigurationException, SecurityException, NoSuchMethodException {
		if (this.classMapper.getTableMap().hasGeneratedField()) {
			final String javaColumnName = this.classMapper.getAllGeneratedColumnJavaNames()[0];
			final IColumn column = this.classMapper.getColumnWithJavaName(javaColumnName);
			final FieldManipulator fieldManipulator = this.manipulators.get(javaColumnName);
			final AColumnValueGenerator generator = this.classMapper.getTableMap().getGeneratorByJavaFieldName(javaColumnName);
			final long[] validValues = generator.getActivateValues();
			this.generatorManipulator = AReflectionManipulatorFactory.getGeneratorManipulator(column, fieldManipulator, validValues);
		}
	}

	@Override
	public T mapRow(String rowNamePrefix, ResultSet rs, int rowNum) {
		final String[] allColumnNames = this.classMapper.getAllColumnJavaNames();
		try {
			final T entity = this.classMapper.getMappedClass().newInstance();
			for (final String columnJavaName : allColumnNames) {
				final FieldManipulator manipulator = this.manipulators.get(columnJavaName);
				manipulator.setValueFromResultSet(entity, rs, rowNamePrefix + manipulator.getColumnName());
			}
			return entity;
		} catch (final Exception e) {
			throw new OrmReflectionException(e);
		}
	}

	@Override
	public Object[] allValues(T entity) {
		return this.getValues(this.classMapper.getAllColumnJavaNames(), entity);
	}

	@Override
	public Object[] allNotGeneratedValues(T entity) {
		return this.getValues(this.classMapper.getAllNotGeneratedColumnJavaNames(), entity);
	}

	@Override
	public Object[] primaryKeyValues(T entity) {
		return this.getValues(this.classMapper.getPrimaryKeyColumnJavaNames(), entity);
	}

	@Override
	public Object[] notPrimaryKeyValues(T entity) {
		return this.getValues(this.classMapper.getNotPrimaryKeyColumnJavaNames(), entity);
	}

	@Override
	public Object[] primaryKeyAndVersionValues(T entity) {
		return this.getValues(this.classMapper.getPrimaryKeyAndVersionColumnJavaNames(), entity);
	}

	@Override
	public void increaseVersion(T entity, boolean firstVersionNumber) {
		if (this.classMapper.getTableMap().isVersionable()){
			try {
				this.versionManipulator.increaseVersion(entity, firstVersionNumber);
			} catch (final Exception e) {
				throw new OrmReflectionException(e);
			}
		}
	}

	@Override
	public void updateGeneratedValues(ResultSet rs, T entity) {
		final String[] allColumnNames = this.classMapper.getAllGeneratedColumnJavaNames();
		try {
			int i=1;
			for (final String columnJavaName : allColumnNames) {
				final FieldManipulator manipulator = this.manipulators.get(columnJavaName);
				manipulator.setValueFromResultSet(entity, rs, i);
				i++;
			}
		} catch (final Exception e) {
			throw new OrmReflectionException(e);
		}
	}

	private Object[] getValues(String[] javaColumnNames, T entity) {
		final Object[] result = new Object[javaColumnNames.length];
		try {
			for (int i=0; i<javaColumnNames.length ; i++) {
				final String javaColumnName = javaColumnNames[i];
				final FieldManipulator manipulator = this.manipulators.get(javaColumnName);
				result[i] = manipulator.getValue(entity);
			}
		} catch (final Exception e) {
			throw new OrmReflectionException(e);
		}
		return result;
	}

	@Override
	public boolean useKeyGenerators(T entity) {
		try {
			return this.generatorManipulator.useGenerator(entity);
		} catch (final Exception e) {
			throw new OrmReflectionException(e);
		}
	}

}
