package com.jpattern.orm.query;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import com.jpattern.orm.classtool.IOrmClassTool;
import com.jpattern.orm.classtool.IOrmClassToolMap;
import com.jpattern.orm.exception.OrmException;
import com.jpattern.orm.exception.OrmNotUniqueResultException;
import com.jpattern.orm.query.clause.From;
import com.jpattern.orm.query.clause.OrderBy;
import com.jpattern.orm.query.clause.OrmFrom;
import com.jpattern.orm.query.clause.OrmOrderBy;
import com.jpattern.orm.query.clause.OrmWhere;
import com.jpattern.orm.query.clause.Where;
import com.jpattern.orm.session.ResultSetReader;
import com.jpattern.orm.session.SessionSqlPerformer;
import com.jpattern.orm.session.SqlPerformer;
import com.jpattern.orm.util.GenericWrapper;

/**
 * 
 * @author Francesco Cina
 *
 * 20/giu/2011
 */
public class OrmFindQuery<T> extends AQuery implements FindQuery<T>, INameSolverConsumer {

	private INameSolver nameSolver = new NullNameSolver();
	private final IOrmClassToolMap ormClassToolMap;
	private final Class<T> clazz;
	private boolean distinct;
	private final SessionSqlPerformer session;
	private int queryTimeout = 0;
	private int maxRows = 0;
	private LockMode lockMode = LockMode.NO_LOCK;
	private final Where<FindQuery<T>> whereExpression = new OrmWhere<FindQuery<T>>(this);
	private final OrmOrderBy<FindQuery<T>> orderBy = new OrmOrderBy<FindQuery<T>>(this);
	private final From<FindQuery<T>> from;

	public OrmFindQuery(IOrmClassToolMap ormClassToolMap, SessionSqlPerformer session, Class<T> clazz) {
		this.ormClassToolMap = ormClassToolMap;
		this.session = session;
		this.clazz = clazz;
		from = new OrmFrom<FindQuery<T>>(ormClassToolMap, this);
	}

	@Override
	public final From<FindQuery<T>> from() throws OrmException {
		return from;
	}

	@Override
	public final Where<FindQuery<T>> where() throws OrmException {
		return whereExpression;
	}

	@Override
	public final OrderBy<FindQuery<T>> orderBy() throws OrmException {
		return orderBy;
	}

	@Override
	public final int getMaxRows() throws OrmException {
		return this.maxRows;
	}

	@Override
	public final FindQuery<T> setQueryTimeout(int queryTimeout) {
		this.queryTimeout = queryTimeout;
		return this;
	}

	@Override
	public final int getQueryTimeout() {
		return this.queryTimeout;
	}

	@Override
	public final FindQuery<T> setMaxRows(int maxRows) throws OrmException {
		this.maxRows = maxRows;
		return this;
	}

	@Override
	public void setNameSolver(INameSolver nameSolver) {
		this.nameSolver = nameSolver;
		where().setNameSolver(nameSolver);
		orderBy().setNameSolver(nameSolver);
		from().setNameSolver(nameSolver);
	}

	@Override
	public List<T> findList() {
		final List<Object> values = new ArrayList<Object>();
		where().appendValues(values);
		final IOrmClassTool<T> ormClassTool = this.ormClassToolMap.getOrmClassTool(this.clazz);

		final ResultSetReader<List<T>> resultSetReader = new ResultSetReader<List<T>>() {

			@Override
			public List<T> read(ResultSet resultSet) throws SQLException {
				final List<T> resultList = new ArrayList<T>();
				int rowCount = 0;
				while ( resultSet.next() ) {
					resultList.add( ormClassTool.getOrmPersistor().mapRow("", resultSet, rowCount++) );
				}
				return resultList;
			}
		};
		final SqlPerformer sqlExec = this.session.sqlPerformer();
		sqlExec.setMaxRows(this.getMaxRows());
		sqlExec.setQueryTimeout(this.getQueryTimeout());
		return sqlExec.query(this.renderSql(), resultSetReader, values.toArray());
	}

	@Override
	public void find(final OrmRowMapper<T> srr) throws OrmException {
		final List<Object> values = new ArrayList<Object>();
		where().appendValues(values);
		final IOrmClassTool<T> ormClassTool = this.ormClassToolMap.getOrmClassTool(this.clazz);

		final ResultSetReader<Object> resultSetReader = new ResultSetReader<Object>() {
			@Override
			public Object read(ResultSet resultSet) throws SQLException {
				int rowCount = 0;
				while ( resultSet.next() ) {
					srr.read( ormClassTool.getOrmPersistor().mapRow("", resultSet, rowCount) , rowCount );
					rowCount++;
				}
				return null;
			}
		};
		final SqlPerformer sqlExec = this.session.sqlPerformer();
		sqlExec.setMaxRows(this.getMaxRows());
		sqlExec.setQueryTimeout(this.getQueryTimeout());
		sqlExec.query(this.renderSql(), resultSetReader, values.toArray());
	}

	@Override
	public T findUnique() throws OrmNotUniqueResultException {
		final List<Object> values = new ArrayList<Object>();
		where().appendValues(values);
		final IOrmClassTool<T> ormClassTool = this.ormClassToolMap.getOrmClassTool(this.clazz);
		final GenericWrapper<Integer> resultsCount = new GenericWrapper<Integer>(null);
		final GenericWrapper<T> wrapper = new GenericWrapper<T>(null);
		final ResultSetReader<Object> resultSetReader = new ResultSetReader<Object>() {
			@Override
			public Object read(ResultSet resultSet) throws SQLException {
				if ( resultSet.next() ) {
					resultsCount.setValue(1);
					wrapper.setValue( ormClassTool.getOrmPersistor().mapRow("", resultSet, 0) );
					if ( resultSet.next() ) {
						resultsCount.setValue(2);
					}
				} else {
					resultsCount.setValue(0);
				}
				return null;
			}
		};
		final SqlPerformer sqlExec = this.session.sqlPerformer();
		sqlExec.setMaxRows(2);
		sqlExec.setQueryTimeout(this.getQueryTimeout());
		sqlExec.query(this.renderSql(), resultSetReader, values.toArray());
		if ( resultsCount.getValue() != 1) {
			if ( resultsCount.getValue() == 0 ) {
				throw new OrmNotUniqueResultException("The query execution returned a number of rows different than one: no results found");
			}
			throw new OrmNotUniqueResultException("The query execution returned a number of rows different than one: more than one result found");
		}
		return wrapper.getValue();
	}

	@Override
	public long findRowCount() {
		final List<Object> values = new ArrayList<Object>();
		where().appendValues(values);
		final SqlPerformer sqlExec = this.session.sqlPerformer();
		sqlExec.setMaxRows(this.getMaxRows());
		sqlExec.setQueryTimeout(this.getQueryTimeout());
		return sqlExec.queryForLong(this.getGeneratedRowCountSql(), values.toArray());
	}

	@Override
	public String getGeneratedRowCountSql() {
		final StringBuilder stringBuilder = new StringBuilder();
		this.renderSelectRowCount(stringBuilder);
		this.renderFrom(stringBuilder);
		this.renderWhere(stringBuilder);
		return stringBuilder.toString();
	}

	@Override
	public FindQuery<T> setDistinct() {
		this.distinct = true;
		return this;
	}

	private void renderSelectRowCount(StringBuilder stringBuilder) {
		stringBuilder.append("SELECT COUNT(*) ");
	}

	@Override
	public final String renderSql() {
		final StringBuilder stringBuilder = new StringBuilder();
		renderSql(stringBuilder);
		return stringBuilder.toString();
	}

	@Override
	public final void renderSql(StringBuilder stringBuilder) {
		renderSelect(stringBuilder);
		renderFrom(stringBuilder);
		renderWhere(stringBuilder);
		renderOrderBy(stringBuilder);
		renderLockMode(stringBuilder);
	}

	private void renderSelect(StringBuilder stringBuilder) {
		final String alias = this.nameSolver.alias(this.clazz);
		stringBuilder.append("SELECT ");
		if (this.distinct) {
			stringBuilder.append("DISTINCT ");
		}
		stringBuilder.append(this.ormClassToolMap.getOrmClassTool(this.clazz).getOrmCRUDQuery().getBaseSelectClause(alias + ".") );
		stringBuilder.append(" ");
	}

	private void renderFrom(StringBuilder stringBuilder) {
		final String alias = this.nameSolver.alias(this.clazz);
		stringBuilder.append("FROM ");
		stringBuilder.append(this.ormClassToolMap.getOrmClassTool(this.clazz).getClassMapper().getTableMap().getTableNameWithSchema() );
		stringBuilder.append( " " );
		stringBuilder.append(alias);
		stringBuilder.append(" ");
		from().renderSql(stringBuilder);
	}

	@Override
	public final void appendValues(List<Object> values) {
		where().appendValues(values);
	}

	private void renderWhere(StringBuilder stringBuilder) {
		where().renderSql(stringBuilder);
	}

	private void renderOrderBy(StringBuilder stringBuilder) {
		orderBy().renderSql(stringBuilder);
	}

	private void renderLockMode(StringBuilder stringBuilder) {
		stringBuilder.append(this.lockMode.getLockMode());
	}

	@Override
	public boolean isDistinct() throws OrmException {
		return this.distinct;
	}

	@Override
	public FindQuery<T> setLockMode(LockMode lockMode) {
		this.lockMode = lockMode;
		return this;
	}

}
